
/*
**
**  $VER: DTConvert.c 1.3 (31.5.97)
**  datatypes.library/Examples/DTConvert
**
**  Converts file into another format using datatypes
**
**  Written 1996/97 by Roland 'Gizzy' Mainz
**
*/

/* amiga includes */
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/libraries.h>
#include <dos/dos.h>
#include <dos/rdargs.h>
#include <dos/dosextens.h>
#include <intuition/classes.h>  /* must be $Id: classes.h,v 40.0 94/02/15 17:46:35 davidj Exp Locker: davidj $ */
#include <intuition/classusr.h>
#include <intuition/cghooks.h>
#include <intuition/icclass.h>
#include <intuition/gadgetclass.h>
#include <datatypes/datatypes.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/textclass.h>
#if 0
#include <datatypes/documentclass.h>
#endif
#include <datatypes/soundclass.h>
#if 0
#include <datatypes/instrumentclass.h>
#endif
#if 0
#include <datatypes/musicclass.h>
#endif
#include <datatypes/pictureclass.h>
#include <datatypes/animationclass.h>

/* amiga prototypes */
#include <clib/dos_protos.h>
#include <clib/exec_protos.h>
#include <clib/intuition_protos.h>
#include <clib/datatypes_protos.h>
#include <clib/dtclass_protos.h>

/* amiga pragmas */
#include <pragmas/dos_pragmas.h>
#include <pragmas/exec_pragmas.h>
#include <pragmas/intuition_pragmas.h>
#include <pragmas/datatypes_pragmas.h>
#include <pragmas/dtclass_pragmas.h>
#include <pragmas/alib_pragmas.h> /* tagcall stubs */

/* ansi includes */
#include <string.h>

/* version string */
#include "DTConvert_rev.h"

/* misc defines */
#ifndef NAME
#define NAME "DTConvert"
#endif /* !NAME */

/*****************************************************************************/

/* SASC specific defines */
#define DISPATCHERFLAGS __saveds __asm
#define REGARGS __asm
#define REGA0 register __a0
#define REGA1 register __a1
#define REGA2 register __a2

#define V( x )      ((VOID *)(x))
#define G( o )      ((struct Gadget *)(o))
#define EXTG( o )   ((struct ExtGadget *)(o))

#define XTAG( expr, tagid ) ((Tag)((expr)?(tagid):(TAG_IGNORE)))

/*****************************************************************************/

__stdargs void kprintf( UBYTE *msg, ... );

#define DODEBUG 1

#ifdef DODEBUG
#define D( x ) x
#else
#define D( x )
#endif /* DODEBUG */

/*****************************************************************************/

/* from <iffp/ilbm.h>*/
/* BMHD flags */

/* Advisory that 8 significant bits-per-gun have been stored in CMAP
 * i.e. that the CMAP is definitely not 4-bit values shifted left.
 * This bit will disable nibble examination by color loading routine.
 */
#define BMHDB_CMAPOK (7U)
#define BMHDF_CMAPOK (1U << BMHDB_CMAPOK)

/*****************************************************************************/

/* Source object for writeanimclass (see ConvertAnimation below) */
#define WRITEANIMA_SourceObject (DTA_UserData)

/*****************************************************************************/

#define TEMPLATE "NAME=SRCNAME/A,"               \
                 "DESTDATATYPE=DATATYPE=DTN,"    \
                 "IFF/S,"                        \
                 "TO=DESTNAME/A"

#define OPT_SRCNAME       (0)
#define OPT_DESTDATATYPE  (1)
#define OPT_IFF           (2)
#define OPT_DESTNAME      (3)
#define OPT_MAX           (4)

/*****************************************************************************/

#ifndef PARAMETERS_STACK
#define PARAMETERS_STACK 1
#define  CLIB_ALIB_PROTOS_H
__stdargs void  NewList( struct List *list );
__stdargs ULONG DoMethodA( Object *obj, Msg message );
__stdargs ULONG DoMethod( Object *obj, unsigned long MethodID, ... );
__stdargs ULONG DoSuperMethodA( struct IClass *cl, Object *obj, Msg message );
__stdargs ULONG DoSuperMethod( struct IClass *cl, Object *obj, unsigned long MethodID, ... );
__stdargs ULONG CoerceMethodA( struct IClass *cl, Object *obj, Msg message );
__stdargs ULONG CoerceMethod( struct IClass *cl, Object *obj, unsigned long MethodID, ... );
__stdargs ULONG SetSuperAttrs( struct IClass *cl, Object *obj, unsigned long Tag1, ... );
#endif /* !PARAMETERS_STACK */

/*****************************************************************************/

extern struct Library *SysBase,
                      *DOSBase;

/*****************************************************************************/

struct Library *IntuitionBase;
struct Library *DataTypesBase;

/*****************************************************************************/

/* version_string */
STRPTR versionstring = VERSTAG;

/*****************************************************************************/

/* local prototypes */
static                 BOOL  ConvertDataType( Object *, struct DataType *, STRPTR );
static                 BOOL  ConvertPicture( Object *, struct DataType *, STRPTR );
static                 BOOL  ConvertAnimation( Object *, struct DataType *, STRPTR );
static DISPATCHERFLAGS ULONG WriteAnimDispatch( REGA0 struct IClass *, REGA2 Object *, REGA1 Msg );
static                 void  mysprintf( STRPTR, STRPTR, ... );

/*****************************************************************************/

/****** DTConvert/DTConvert **************************************************
*
*    NAME
*       DTConvert -- DataTypes-based conversion tool
*
*    FORMAT
*       DTConvert [NAME|SRCNAME <file>] [DESTDATATYPE|DATATYPE|DTN <datatype>]
*                 [IFF] [TO|DESTNAME <file>]
*
*    TEMPLATE
*       NAME=SRCNAME/A,DESTDATATYPE=DATATYPE=DTN,IFF/S,TO=DESTNAME/A
*
*    PURPOSE
*       Convert data from one format into another format.
*
*    DESCRIPTION
*       DTConvert converts data from one format into another format,
*       e.g. picture -> picture,
*            animation -> anmation,
*            movie -> movie,
*            sound -> sound (n/a),
*            text -> text (n/a)
*
*      You simply have to set the souce file name (SRCNAME), the
*      datatype to convert to (DATATYPE) (or the IFF switch if you want to 
*      convert into the base-IFF format and the destination file name (TO).
*
*    BUGS
*
*    NOTES
*
*    HISTORY
*       V1.1:
*         - First release to dta@amigawolrd.com mailinglist.
*
*       V1.2:
*         - Added support for animation.datatype subclasses
*
*       V1.3:
*         - Minor fixes.
*
*    AUTHOR's REQUEST
*        By  releasing  this program I do  not  place any obligations on you,
*        feel free to share this program with your  friends (and enemies).
*
*        If you want to blame me, report any bugs, or wants a new version
*        send your letter to:
*                        Roland Mainz
*                        Hohenstaufenstrae 8
*                        52388 Nrvenich
*                        GERMANY
*
*        Phone: (+49)(0)2426/901568
*
*        EMAIL is also available:
*        GISBURN@w-specht.rhein-ruhr.de
*
*        If you want to send me attachments larger than 1MB (up to 5MB,
*        more with my permission):
*        Up to July 1997 I'm reachable using this email address, too:
*        Reinhold.A.Mainz@KBV.DE
*
*        | Please put your name and address in your mails !
*        | German mailers should add their phone numbers.
*        | See BUGS section above when submitting bug reports.
*
*        Sorry, but I can only look once a week for mails.
*        If you don't hear something from me within three weeks, please
*        send your mail again (but watch about new releases) (problems with
*        this email port are caused by reconfigurations, hackers, network
*        problems etc.).
*
*        The  entire  "DTConvert"  package may  be  noncommercially
*        redistributed, provided  that  the package  is always  distributed
*        in it's complete  form (including it's documentation).  A small copy
*        fee for media costs is okay but any kind of commercial distribution
*        is strictly forbidden! Comments  and  suggestions  how  to  improve
*        this program  are generally appreciated!
*
*        Thanks to Matt Dillon for his DICE, David Junod for this datatypes
*        environment and Olaf 'Olsen' Barthel for his help, ideas and some 
*        text clips from his documentations.
*
*    SEE ALSO
*
******************************************************************************
*
*/


void main( void )
{
    LONG           options[ OPT_MAX ];
    struct RDArgs *rdargs;

    if( (SysBase -> lib_Version) < 39UL )
    {
      Printf( "requires 3.x or beyond to run\n" );
    }
    else
    {
      /* Clear the option array */
      memset( options, 0, sizeof( options ) );

      /* Parse the arguments */
      if( rdargs = ReadArgs( TEMPLATE, options, NULL ) )
      {
        /* Open the libraries... */
        if( IntuitionBase = OpenLibrary( "intuition.library", 39UL ) )
        {
          if( DataTypesBase = OpenLibrary( "datatypes.library", 45UL ) )
          {
            struct DataType *destdtn = NULL;

            if( options[ OPT_DESTDATATYPE ] )
            {
              destdtn = ObtainDataTypeA( DTST_RAM, (APTR)options[ OPT_DESTDATATYPE ], NULL );
            }

            if( destdtn || (options[ OPT_IFF ] && (!options[ OPT_DESTDATATYPE ])) )
            {
              Object *srcdto;

              Printf( "loading source data \"%s\"\n", (STRPTR)options[ OPT_SRCNAME ] );

              if( srcdto = NewDTObject( (APTR)options[ OPT_SRCNAME ], XTAG( destdtn, DTA_GroupID ), ((destdtn)?(destdtn -> dtn_Header -> dth_GroupID):(0UL)),
                                                                      TAG_DONE ) )
              {
                if( ConvertDataType( srcdto, destdtn, (STRPTR)options[ OPT_DESTNAME ] ) )
                {
                  Printf( "file successfully converted\n" );
                }
                else
                {
                  Printf( "eror %ld\n", IoErr() );
                }

                DisposeDTObject( srcdto );
              }
              else
              {
                Printf( "can't create destination object\n" );
              }

              ReleaseDataType( destdtn );
            }
            else
            {
              Printf( "can' open destination datatype\n" );
            }

            if( IoErr() )
            {
              PrintFault( IoErr(), NULL );
            }

            CloseLibrary( DataTypesBase );
          }
          else
          {
            Printf( "Couldn't open datatypes.library V45\n" );
          }

          CloseLibrary( IntuitionBase );
        }
        else
        {
          Printf( "Couldn't open intuition.library V39\n" );
        }

        /* Free the allocated memory after ReadArgs */
        FreeArgs( rdargs );
      }
      else
      {
        /* ReadArgs failed */
        PrintFault( IoErr(), "DTConvert" );
      }
    }
}


static
BOOL ConvertDataType( Object *srcdto, struct DataType *destdtn, STRPTR destfile )
{
    BOOL success = FALSE;
    LONG error   = 0L;

    if( srcdto && destfile )
    {
      struct DataType *srcdtn;

      GetDTAttrs( srcdto, DTA_DataType, (&srcdtn), TAG_DONE );

      Printf( "converting \"%s\" %s into \"%s\" %s\n",
              (srcdtn  -> dtn_Header -> dth_Name),
              GetDTString( (srcdtn -> dtn_Header -> dth_GroupID) ),
              ((destdtn)?(destdtn -> dtn_Header -> dth_Name):("IFF")),
              ((destdtn)?GetDTString( (destdtn -> dtn_Header -> dth_GroupID) ):("IFF")) );

      switch( srcdtn -> dtn_Header -> dth_GroupID )
      {
        case GID_SYSTEM:
        {
            /* There is currently no system.datatype base class defined */
            error = ERROR_ACTION_NOT_KNOWN;
        }
            break;

        case GID_TEXT:
        {
            error = ERROR_NOT_IMPLEMENTED;
        }
            break;

        case GID_DOCUMENT:
        {
            /* There is currently no document.datatype base class defined */
            error = ERROR_ACTION_NOT_KNOWN;
        }
            break;

        case GID_SOUND:
        {
            error = ERROR_NOT_IMPLEMENTED;
        }
            break;

        case GID_INSTRUMENT:
        {
            /* There is currently no instrument.datatype base class defined */
            error = ERROR_ACTION_NOT_KNOWN;
        }
            break;

        case GID_MUSIC:
        {
            /* There is currently no music.datatype base class defined */
            error = ERROR_ACTION_NOT_KNOWN;
        }
            break;

        case GID_PICTURE:
        {
            success = ConvertPicture( srcdto, destdtn, destfile );
            error = IoErr();
        }
            break;

        case GID_ANIMATION:
        case GID_MOVIE:
        {
            success = ConvertAnimation( srcdto, destdtn, destfile );
            error = IoErr();
        }
            break;

        default:
        {
            error = ERROR_ACTION_NOT_KNOWN;
        }
            break;
      }
    }

    SetIoErr( error );

    return( success );
}


static
BOOL ConvertPicture( Object *srcdto, struct DataType *destdtn, STRPTR destfile )
{
    BOOL success = FALSE;
    LONG error   = 0L;

    Printf( "picture converter\n" );

    if( srcdto && destfile )
    {
      Object          *destdto;
      struct DataType *srcdtn;
      struct BitMap   *bm;

      /* Some (ugly) picture.datatype needs a DTM_PROCLAYOUT before
       * allowing to access their bitmaps
       */
      GetDTAttrs( srcdto, PDTA_BitMap, (&bm), TAG_DONE );

      if( bm == NULL )
      {
        struct gpLayout  gpl;

        Printf( "requires layouting...\n" );

        /* "Layout" picture (for those picture classes which need this,
         * __NOT__ recommened)
         */
        SetDTAttrs( srcdto, NULL, NULL, PDTA_Remap, FALSE, TAG_DONE );

        gpl . MethodID    = DTM_PROCLAYOUT;
        gpl . gpl_GInfo   = NULL;
        gpl . gpl_Initial = 1L;

        DoDTMethodA( srcdto, NULL, NULL, (Msg)(&gpl) );
      }

      /* Get source datatype, see below */
      GetDTAttrs( srcdto, DTA_DataType, (&srcdtn), TAG_DONE );

      /* Create empty picture.datatype subclass object */
      if( destdto = NewDTObject( NULL, DTA_SourceType, DTST_RAM,
                                       XTAG( destdtn,    DTA_DataType ), destdtn,
                                       XTAG( (!destdtn), DTA_GroupID  ), (srcdtn -> dtn_Header -> dth_GroupID),
                                       TAG_DONE ) )
      {
        STRPTR                objname,
                              objauthor,
                              objannotation,
                              objcopyright,
                              objversion;
        ULONG                 modeid;
        ULONG                 numcolors;
        struct BitMapHeader  *sbmh,
                             *dbmh;
        ULONG                *scregs;
        struct ColorRegister *scm;

        if( GetDTAttrs( srcdto, DTA_ObjName,         (&objname),
                                DTA_ObjAuthor,       (&objauthor),
                                DTA_ObjAnnotation,   (&objannotation),
                                DTA_ObjCopyright,    (&objcopyright),
                                DTA_ObjVersion,      (&objversion),
                                PDTA_ModeID,         (&modeid),
                                PDTA_BitMapHeader,   (&sbmh),
                                PDTA_NumColors,      (&numcolors),
                                PDTA_CRegs,          (&scregs),
                                PDTA_ColorRegisters, (&scm),
                                PDTA_BitMap,         (&bm),
                                TAG_DONE ) == 11UL )
        {
          if( GetDTAttrs( destdto, PDTA_BitMapHeader, (&dbmh), TAG_DONE ) )
          {
            if( sbmh && bm && dbmh )
            {
              ULONG                *dcregs;
              struct ColorRegister *dcm;

              /* copy bitmapheader (RAW) from animation bmh... */
              *dbmh = *sbmh;

              /* ...and fill in attributes which may be different */
              dbmh -> bmh_Compression = cmpNone;

              /* we're writing a full palette, set this flag (currently NOP, but...) */
              dbmh -> bmh_Pad    = BMHDF_CMAPOK; /* also known as bmh_Flags */

              /* Set datas "name", "author", "annotation", "copyright", "version" etc.,
               * screen mode id,
               * our bitmap and
               * the requested number of colors...
               */
              SetDTAttrs( destdto, NULL, NULL,
                          DTA_ObjName,       objname,
                          DTA_ObjAuthor,     objauthor,
                          DTA_ObjAnnotation, objannotation,
                          DTA_ObjCopyright,  objcopyright,
                          DTA_ObjVersion,    objversion,
                          PDTA_ModeID,       modeid,
                          PDTA_NumColors,    numcolors,
                          PDTA_BitMap,       bm,
                          TAG_DONE );

              /* Some (ugly) picture.datatype needs a DTM_PROCLAYOUT before
               * allowing to access their bitmaps
               */
              GetDTAttrs( destdto, PDTA_BitMap, (&bm), TAG_DONE );

              if( bm == NULL )
              {
                struct gpLayout  gpl;

                Printf( "destination requires layouting...\n" );

                /* "Layout" picture (for those picture classes which need this,
                 * __NOT__ recommened)
                 */
                SetDTAttrs( destdto, NULL, NULL, PDTA_Remap, FALSE, TAG_DONE );

                gpl . MethodID    = DTM_PROCLAYOUT;
                gpl . gpl_GInfo   = NULL;
                gpl . gpl_Initial = 1L;

                DoDTMethodA( destdto, NULL, NULL, (Msg)(&gpl) );
              }

              /* Get color tables (dest cregs, dest cm)  */
              if( GetDTAttrs( destdto, PDTA_CRegs,          (&dcregs),
                                       PDTA_ColorRegisters, (&dcm),
                                       TAG_DONE ) == 2UL )
              {
                /* Success ? */
                if( dcregs && dcm )
                {
                  ULONG i;

                  /* Copy color tables... */
                  for( i = 0UL ; i < numcolors ; i++ )
                  {
                    dcregs[ ((i * 3) + 0) ] = scregs[ ((i * 3) + 0) ]; /* copy r, */
                    dcregs[ ((i * 3) + 1) ] = scregs[ ((i * 3) + 1) ]; /*      g, */
                    dcregs[ ((i * 3) + 2) ] = scregs[ ((i * 3) + 2) ]; /*      b  */

                    *dcm++ = *scm++; /* copy colorregisters */
                  }

                  Printf( "saving picture: \"%s\"\n", destfile );
                  if( !(success = SaveDTObjectA( destdto, NULL, NULL, destfile, DTWM_RAW, TRUE, NULL )) )
                  {
                    Printf( "saving failed\n" );

                    error = IoErr();
                  }
                }
                else
                {
                  Printf( "picture object mem err (no color tables)\n" );
                  error = ERROR_NO_FREE_STORE;
                }
              }
              else
              {
                /* can't get required args from object */
                error = ERROR_OBJECT_WRONG_TYPE;
              }

              /* src datatype will free it's bitmap  */
              SetDTAttrs( destdto, NULL, NULL, PDTA_BitMap, NULL, TAG_DONE );
            }
            else
            {
              Printf( "can't get bitmap\n" );
              error = ERROR_NO_FREE_STORE;
            }
          }
          else
          {
            /* can't get required args from object */
            Printf( "can't get PDTA_BitMapHeader\n" );
            error = ERROR_OBJECT_WRONG_TYPE;
          }
        }

        /* Get rid of the picture object */
        DisposeDTObject( destdto );
      }
      else
      {
        Printf( "can't create empty object\n" );

        /* NewDTObject failed */
        error = IoErr();
      }
    }

    SetIoErr( error );

    return( success );
}


static
BOOL ConvertAnimation( Object *srcdto, struct DataType *destdtn, STRPTR destfile )
{
    BOOL success = FALSE;
    LONG error   = 0L;

    Printf( "animation converter\n" );

    if( srcdto && destfile && destdtn )
    {
      TEXT             classname[ 256 ];
      struct Library  *DTClassBase;
      Object          *destdto;
      struct DataType *srcdtn;

      /* Get source datatype, see below */
      GetDTAttrs( srcdto, DTA_DataType, (&srcdtn), TAG_DONE );

      /* Open class library... */
      mysprintf( classname, "DataTypes/%s.datatype", (destdtn -> dtn_Header -> dth_BaseName) );

      if( DTClassBase = OpenLibrary( classname, 0UL ) )
      {
        struct IClass *AnimClass;

        if( AnimClass = ObtainEngine() )
        {
          struct IClass *WriteAnimClass;

          /* Create our new class... */
          if( WriteAnimClass = MakeClass( NULL, NULL, AnimClass, 0UL, 0UL ) )
          {
            /* Prepare class for usage... */
            WriteAnimClass -> cl_Dispatcher . h_Entry = (HOOKFUNC)WriteAnimDispatch;
            WriteAnimClass -> cl_UserData             = 0UL; /* datatypes.library expects here a (struct Library *) or NULL,
                                                              * therefore we cannot use this for our context data
                                                              */

            /* Create empty subclass object (from our dest write class) */
            if( destdto = (Object *)NewDTObject( NULL, DTA_SourceType,          DTST_RAM,
                                                       DTA_DataType,            destdtn,
                                                       DTA_Class,               WriteAnimClass,
                                                       WRITEANIMA_SourceObject, srcdto,
                                                       TAG_DONE ) )
            {
              STRPTR                objname,
                                    objauthor,
                                    objannotation,
                                    objcopyright,
                                    objversion;
              ULONG                 modeid;
              ULONG                 numcolors;
              struct BitMap        *bm;
              struct BitMapHeader  *sbmh,
                                   *dbmh;
              ULONG                *scregs;
              struct ColorRegister *scm;
              ULONG                 width,
                                    height,
                                    depth;
              ULONG                 frames,
                                    fps,
                                    finc;
              BYTE                 *sample;
              ULONG                 samplelength,
                                    period,
                                    volume,
                                    cycles;
                                    
              if( GetDTAttrs( srcdto, DTA_ObjName,          (&objname),
                                      DTA_ObjAuthor,        (&objauthor),
                                      DTA_ObjAnnotation,    (&objannotation),
                                      DTA_ObjCopyright,     (&objcopyright),
                                      DTA_ObjVersion,       (&objversion),
                                      PDTA_BitMapHeader,    (&sbmh),
                                      ADTA_ModeID,          (&modeid),
                                      ADTA_NumColors,       (&numcolors),
                                      ADTA_CRegs,           (&scregs),
                                      ADTA_ColorRegisters,  (&scm),
                                      ADTA_KeyFrame,        (&bm),
                                      ADTA_Width,           (&width),
                                      ADTA_Height,          (&height),
                                      ADTA_Depth,           (&depth),
                                      ADTA_Frames,          (&frames),
                                      ADTA_FramesPerSecond, (&fps),
                                      ADTA_FrameIncrement,  (&finc),
                                      ADTA_Sample,          (&sample),
                                      ADTA_SampleLength,    (&samplelength),
                                      ADTA_Period,          (&period),
                                      ADTA_Volume,          (&volume),
                                      ADTA_Cycles,          (&cycles),
                                      TAG_DONE ) == 22UL )
              {
                if( GetDTAttrs( destdto, PDTA_BitMapHeader, (&dbmh), TAG_DONE ) )
                {
                  if( sbmh && bm && dbmh )
                  {
                    ULONG                *dcregs;
                    struct ColorRegister *dcm;

                    /* copy bitmapheader (RAW) from animation bmh... */
                    *dbmh = *sbmh;

                    /* ...and fill in attributes which may be different */
                    dbmh -> bmh_Compression = cmpNone;

                    /* we're writing a full palette, set this flag (currently NOP, but...) */
                    dbmh -> bmh_Pad    = BMHDF_CMAPOK; /* also known as bmh_Flags */

                    /* Set datas "name", "author", "annotation", "copyright", "version" etc.,
                     * screen mode id,
                     * our bitmap and
                     * the requested number of colors...
                     */
                    SetDTAttrs( destdto, NULL, NULL,
                                DTA_ObjName,          objname,
                                DTA_ObjAuthor,        objauthor,
                                DTA_ObjAnnotation,    objannotation,
                                DTA_ObjCopyright,     objcopyright,
                                DTA_ObjVersion,       objversion,
                                PDTA_BitMapHeader,    sbmh,
                                ADTA_ModeID,          modeid,
                                ADTA_NumColors,       numcolors,
                                ADTA_CRegs,           scregs,
                                ADTA_ColorRegisters,  scm,
                                ADTA_KeyFrame,        bm,
                                ADTA_Width,           width,
                                ADTA_Height,          height,
                                ADTA_Depth,           depth,
                                ADTA_Frames,          frames,
                                ADTA_FramesPerSecond, fps,
                                ADTA_FrameIncrement,  finc,
                                ADTA_Sample,          sample,
                                ADTA_SampleLength,    samplelength,
                                ADTA_Period,          period,
                                ADTA_Volume,          volume,
                                ADTA_Cycles,          cycles,
                                TAG_DONE );

                    /* Get color tables (dest cregs, dest cm)  */
                    if( GetDTAttrs( destdto, ADTA_CRegs,          (&dcregs),
                                             ADTA_ColorRegisters, (&dcm),
                                             TAG_DONE ) == 2UL )
                    {
                      /* Success ? */
                      if( dcregs && dcm )
                      {
                        ULONG i;

                        /* Copy color tables... */
                        for( i = 0UL ; i < numcolors ; i++ )
                        {
                          dcregs[ ((i * 3) + 0) ] = scregs[ ((i * 3) + 0) ]; /* copy r, */
                          dcregs[ ((i * 3) + 1) ] = scregs[ ((i * 3) + 1) ]; /*      g, */
                          dcregs[ ((i * 3) + 2) ] = scregs[ ((i * 3) + 2) ]; /*      b  */

                          *dcm++ = *scm++; /* copy colorregisters */
                        }

                        Printf( "saving animation: \"%s\"\n", destfile );
                        if( !(success = SaveDTObjectA( destdto, NULL, NULL, destfile, DTWM_RAW, TRUE, NULL )) )
                        {
                          Printf( "saving failed\n" );

                          error = IoErr();
                        }
                      }
                      else
                      {
                        Printf( "picture object mem err (no color tables)\n" );
                        error = ERROR_NO_FREE_STORE;
                      }
                    }
                    else
                    {
                      /* can't get required args from object */
                      error = ERROR_OBJECT_WRONG_TYPE;
                    }

                    /* src datatype will free it's bitmap */
                    SetDTAttrs( destdto, NULL, NULL, ADTA_KeyFrame, NULL, TAG_DONE );
                  }
                  else
                  {
                    Printf( "can't get bitmap\n" );
                    error = ERROR_NO_FREE_STORE;
                  }
                }
                else
                {
                  /* can't get required args from object */
                  Printf( "can't get PDTA_BitMapHeader\n" );
                  error = ERROR_OBJECT_WRONG_TYPE;
                }
              }

              /* Get rid of the picture object */
              DisposeDTObject( destdto );
            }
            else
            {
              Printf( "can't create empty object\n" );

              /* NewDTObject failed */
              error = IoErr();
            }

            (void)FreeClass( WriteAnimClass );
          }
        }
        else
        {
          Printf( "can't obtain output anim class\n" );
        }
        
        CloseLibrary( DTClassBase );
      }
    }

    SetIoErr( error );

    return( success );
}


/* WriteAnimClass dispatcher */
static
DISPATCHERFLAGS
ULONG WriteAnimDispatch( REGA0 struct IClass *cl, REGA2 Object *o, REGA1 Msg msg )
{
    ULONG retval = 0UL;

    switch( msg -> MethodID )
    {
        case OM_NEW:
        {
            /* Assumes that WRITEANIMA_SourceObject contains a valid object
             * (WRITEANIMA_SourceObject s mapped to DTA_UserData; datatypesclass handles
             * storage of this attribute :-)
             */
            if( retval = DoSuperMethodA( cl, o, msg ) )
            {
              /* NOP */
            }
        }
            break;

        case OM_DISPOSE:
        {
            DoSuperMethodA( cl, o, msg );
        }
            break;

        case OM_UPDATE:
        {
            /* Avoid OM_NOTIFY loops */
            if( DoMethod( o, ICM_CHECKLOOP ) )
            {
              break;
            }
        }
        case OM_SET:
        {
            /* Pass the attributes to the text class and force a refresh
             * if we need it
             */
            if( retval = DoSuperMethodA( cl, o, msg ) )
            {
/* writeanimclass is alwasy the top instance... */
#ifdef COMMENTED_OUT
              /* Top instance ? */
              if( OCLASS( o ) == cl )
#endif /* COMMENTED_OUT */
              {
                struct RastPort *rp;

                /* Get a pointer to the rastport */
                if( rp = ObtainGIRPort( (((struct opSet *)msg) -> ops_GInfo) ) )
                {
                  struct gpRender gpr;

                  /* Force a redraw */
                  gpr . MethodID   = GM_RENDER;
                  gpr . gpr_GInfo  = ((struct opSet *)msg) -> ops_GInfo;
                  gpr . gpr_RPort  = rp;
                  gpr . gpr_Redraw = GREDRAW_UPDATE;

                  DoMethodA( o, (Msg)(&gpr) );

                  /* Release the temporary rastport */
                  ReleaseGIRPort( rp );
                }

                retval = 0UL;
              }
            }
        }
            break;

        /* This two methods are superset by writeanimclass (e.g. us) */
        case ADTM_LOADFRAME:
        case ADTM_UNLOADFRAME:
        {
            Object *source;
            
            if( GetDTAttrs( o, WRITEANIMA_SourceObject, (&source), TAG_DONE ) )
            {
              if( source )
              {
                /* Route msg to source object... */
                retval = DoMethodA( source, msg );
              }
            }
        }
            break;

        /* Let the superclass handle everything else */
        default:
        {
            retval = DoSuperMethodA( cl, o, msg );
        }
            break;
    }

    return( retval );
}


static
void mysprintf( STRPTR buffer, STRPTR fmt, ... )
{
    APTR args;

    args = (APTR)((&fmt) + 1);

    RawDoFmt( fmt, args, (void (*))"\x16\xc0\x4e\x75", buffer );
}


